# Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

"""
:mod:`mdx_partial_gfm` -- Partial extension for GFM (READMEs, wiki)
===================================================================

An extension that is as compatible as possible with GitHub-flavored
Markdown (GFM).

This extension aims to be compatible with the variant of GFM that GitHub
uses for Markdown-formatted gists and files (including READMEs). This
variant seems to have all the extensions described in the `GFM
documentation`_, except:

- Newlines in paragraphs are not transformed into ``br`` tags.
- Intra-GitHub links to commits, repositories, and issues are not
  supported.

If you need support for features specific to GitHub comments and issues,
please use :class:`mdx_gfm.GithubFlavoredMarkdownExtension`.

.. _GFM documentation: https://guides.github.com/features/mastering-markdown/
"""

from markdown.extensions import Extension
from markdown.extensions.tables import TableExtension

import gfm


def makeExtension(**kwargs):
    return PartialGithubFlavoredMarkdownExtension(**kwargs)


class PartialGithubFlavoredMarkdownExtension(Extension):
    """
    An extension that is as compatible as possible with GitHub-flavored
    Markdown (GFM).

    This extension aims to be compatible with the variant of GFM that GitHub
    uses for Markdown-formatted gists and files (including READMEs). This
    variant seems to have all the extensions described in the `GFM
    documentation`_, except:

    - Newlines in paragraphs are not transformed into ``br`` tags.
    - Intra-GitHub links to commits, repositories, and issues are not
      supported.

    If you need support for features specific to GitHub comments and issues,
    please use :class:`mdx_gfm.GithubFlavoredMarkdownExtension`.

    .. _GFM documentation: https://guides.github.com/features/mastering-markdown/
    """

    def extendMarkdown(self, md):
        # Built-in extensions
        TableExtension().extendMarkdown(md)

        # Custom extensions
        gfm.AutolinkExtension().extendMarkdown(md)
        gfm.AutomailExtension().extendMarkdown(md)
        gfm.SemiSaneListExtension().extendMarkdown(md)
        gfm.StandaloneFencedCodeExtension().extendMarkdown(md)
        gfm.StrikethroughExtension().extendMarkdown(md)
        gfm.TaskListExtension().extendMarkdown(md)



# Copyright (c) 2016, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

"""
:mod:`gfm.tasklist` -- Task list support
========================================

The :mod:`gfm.tasklist` module provides GitHub-like support for task lists.
Those are normal lists with a checkbox-like syntax at the beginning of items
that will be converted to actual checkbox inputs. Nested lists are supported.

Example syntax::

   - [x] milk
   - [ ] eggs
   - [x] chocolate
   - [ ] if possible:
       1. [ ] solve world peace
       2. [ ] solve world hunger

.. NOTE::
   GitHub has support for updating the Markdown source text by toggling the
   checkbox (by clicking on it). This is not supported by this extension, as it
   requires server-side processing that is out of scope of a Markdown parser.

Available configuration options
-------------------------------

================== ============== ======== ===========
Name               Type           Default  Description
================== ============== ======== ===========
``unordered``      bool           ``True`` Set to ``False`` to disable parsing of unordered lists.
``ordered``        bool           ``True`` Set to ``False`` to disable parsing of ordered lists.
``max_depth``      integer        ∞        Set to a positive integer to stop parsing nested task
                                           lists that are deeper than this limit.
``list_attrs``     dict, callable ``{}``   Attributes to be added to the ``<ul>`` or ``<ol>`` element
                                           containing the items.
``item_attrs``     dict, callable ``{}``   Attributes to be added to the ``<li>`` element containing
                                           the checkbox. See `Item attributes`_.
``checkbox_attrs`` dict, callable ``{}``   Attributes to be added to the checkbox element.
                                           See `Checkbox attributes`_.
================== ============== ======== ===========

List attributes
***************

If option ``list_attrs`` is a *dict*, the key-value pairs will be applied to
the ``<ul>`` (resp. ``<ol>``) unordered (resp. ordered) list element, that is
the parent element of the ``<li>`` elements.

.. warning::

   These attributes are applied to all nesting levels of lists, that is,
   to both the root lists and their potential sub-lists, recursively.

   You can control this behavior by using a *callable* instead (see below).

If option ``list_attrs`` is a *callable*, it should be a function that respects
the following prototype::

   def function(list, depth: int) -> dict:

where:

- ``list`` is the ``<ul>`` or ``<ol>`` element;
- ``depth`` is the depth of this list relative to its root list (root lists have
  a depth of 1).

The returned *dict* items will be applied as HTML attributes to the list
element.

.. note::

   Thanks to this feature, you could apply attributes to root lists only.
   Take this code sample::

      import markdown
      from gfm import TaskListExtension

      def list_attr_cb(list, depth):
          if depth == 1:
              return {'class': 'tasklist'}
          return {}

      tl_ext = TaskListExtension(list_attrs=list_attr_cb)

      print(markdown.markdown(\"""
      - [x] some thing
      - [ ] some other
          - [ ] sub thing
          - [ ] sub other
      \""", extensions=[tl_ext]))

   In this example, only the root list will have the ``tasklist`` class
   attribute, not the one containing “sub” items.


Item attributes
***************

If option ``item_attrs`` is a *dict*, the key-value pairs will be applied to
the ``<li>`` element as its HTML attributes.

Example::

   item_attrs={'class': 'list-item'}

will result in::

   <li class="list-item">...</li>

If option ``item_attrs`` is a *callable*, it should be a function that
respects the following prototype::

   def function(parent, element, checkbox) -> dict:

where:

- ``parent`` is the ``<li>`` parent element;
- ``element`` is the ``<li>`` element;
- ``checkbox`` is the generated ``<input type="checkbox">`` element.

The returned *dict* items will be applied as HTML attributes to the ``<li>``
element containing the checkbox.

Checkbox attributes
*******************

If option ``checkbox_attrs`` is a *dict*, the key-value pairs will be applied to
the ``<input type="checkbox">`` element as its HTML attributes.

Example::

   checkbox_attrs={'class': 'list-cb'}

will result in::

   <li><input type="checkbox" class="list-cb"> ...</li>

If option ``checkbox_attrs`` is a *callable*, it should be a function that
respects the following prototype::

   def function(parent, element) -> dict:

where:

- ``parent`` is the ``<li>`` parent element;
- ``element`` is the ``<li>`` element.

The returned *dict* items will be applied as HTML attributes to the checkbox
element.

Typical usage
-------------

.. testcode::

   import markdown
   from gfm import TaskListExtension

   print(markdown.markdown(\"""
   - [x] milk
   - [ ] eggs
   - [x] chocolate
   - not a checkbox
   \""", extensions=[TaskListExtension()]))

.. testoutput::

   <ul>
   <li><input checked="checked" disabled="disabled" type="checkbox" /> milk</li>
   <li><input disabled="disabled" type="checkbox" /> eggs</li>
   <li><input checked="checked" disabled="disabled" type="checkbox" /> chocolate</li>
   <li>not a checkbox</li>
   </ul>

"""

import markdown
from functools import reduce
from markdown.treeprocessors import Treeprocessor
import xml.etree.ElementTree as etree


def _to_list(obj):
    if isinstance(obj, str):
        return [obj]
    return list(obj)


class TaskListProcessor(Treeprocessor):
    def __init__(self, ext):
        super(TaskListProcessor, self).__init__()
        self.ext = ext

    def run(self, root):
        ordered = self.ext.getConfig("ordered")
        unordered = self.ext.getConfig("unordered")
        if not ordered and not unordered:
            return root

        checked_pattern = _to_list(self.ext.getConfig("checked"))
        unchecked_pattern = _to_list(self.ext.getConfig("unchecked"))
        if not checked_pattern and not unchecked_pattern:
            return root

        prefix_length = reduce(
            max, (len(e) for e in checked_pattern + unchecked_pattern)
        )

        item_attrs = self.ext.getConfig("item_attrs")
        list_attrs = self.ext.getConfig("list_attrs")
        base_cb_attrs = self.ext.getConfig("checkbox_attrs")
        max_depth = self.ext.getConfig("max_depth")
        if not max_depth:
            max_depth = float("inf")

        lists = set()
        stack = [(root, None, 0)]

        while stack:
            el, parent, depth = stack.pop()

            if (
                parent
                and el.tag == "li"
                and (parent.tag == "ul" and unordered or parent.tag == "ol" and ordered)
            ):
                depth += 1
                text = (el.text or "").lstrip()
                lower_text = text[:prefix_length].lower()
                found = False
                checked = False

                for p in checked_pattern:
                    if lower_text.startswith(p):
                        found = True
                        checked = True
                        text = text[len(p) :]
                        break
                else:
                    for p in unchecked_pattern:
                        if lower_text.startswith(p):
                            found = True
                            text = text[len(p) :]
                            break

                if found:
                    # Add root <ol> or <ul> element to the list set
                    if list_attrs:
                        lists.add((parent, depth))

                    # Checkbox attributes
                    attrs = {"type": "checkbox", "disabled": "disabled"}
                    if checked:
                        attrs["checked"] = "checked"
                    # Give user a chance to update checkbox attributes
                    attrs.update(
                        base_cb_attrs(parent, el)
                        if callable(base_cb_attrs)
                        else base_cb_attrs
                    )
                    checkbox = etree.Element("input", attrs)
                    checkbox.tail = text
                    # Prepend checkbox to <li>
                    el.text = ""
                    el.insert(0, checkbox)
                    # Give user a chance to update <li> attributes
                    for k, v in (
                        item_attrs(parent, el, checkbox)
                        if callable(item_attrs)
                        else item_attrs
                    ).items():
                        el.set(k, v)

            if depth < max_depth:
                for child in el:
                    stack.append((child, el, depth))

        for list, depth in lists:
            for k, v in (
                list_attrs(list, depth) if callable(list_attrs) else list_attrs
            ).items():
                list.set(k, v)

        return root


class TaskListExtension(markdown.Extension):
    """
    An extension that supports GitHub task lists. Both ordered and unordered
    lists are supported and can be separately enabled. Nested lists are
    supported.

    Example::

       - [x] milk
       - [ ] eggs
       - [x] chocolate
       - [ ] if possible:
           1. [ ] solve world peace
           2. [ ] solve world hunger
    """

    def __init__(self, **kwargs):
        self.config = {
            "ordered": [True, "Enable parsing of ordered lists"],
            "unordered": [True, "Enable parsing of unordered lists"],
            "checked": ["[x]", "The checked state pattern"],
            "unchecked": ["[ ]", "The unchecked state pattern"],
            "max_depth": [0, "Maximum list nesting depth (None for " "unlimited)"],
            "list_attrs": [
                {},
                "Additional attribute dict (or callable) to "
                "add to the <ul> or <ol> element",
            ],
            "item_attrs": [
                {},
                "Additional attribute dict (or callable) to " "add to the <li> element",
            ],
            "checkbox_attrs": [
                {},
                "Additional attribute dict (or callable) "
                "to add to the checkbox <input> element",
            ],
        }
        super().__init__(**kwargs)

    def extendMarkdown(self, md):
        md.treeprocessors.register(TaskListProcessor(self), "gfm-tasklist", 100)



# Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

"""
:mod:`mdx_gfm` -- Full extension for GFM (comments, issues)
===========================================================

An extension that is as compatible as possible with GitHub-flavored
Markdown (GFM).

This extension aims to be compatible with the standard GFM that GitHub uses
for comments and issues. It has all the extensions described in the `GFM
documentation`_, except for intra-GitHub links to commits, repositories,
and issues.

Note that Markdown-formatted gists and files (including READMEs) on GitHub
use a slightly different variant of GFM. For that, use
:class:`mdx_partial_gfm.PartialGithubFlavoredMarkdownExtension`.

.. _GFM documentation: https://guides.github.com/features/mastering-markdown/
"""

# Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

from markdown.extensions.nl2br import Nl2BrExtension

from mdx_partial_gfm import PartialGithubFlavoredMarkdownExtension


def makeExtension(**kwargs):
    return GithubFlavoredMarkdownExtension(**kwargs)


class GithubFlavoredMarkdownExtension(PartialGithubFlavoredMarkdownExtension):
    """
    An extension that is as compatible as possible with GitHub-flavored
    Markdown (GFM).

    This extension aims to be compatible with the standard GFM that GitHub uses
    for comments and issues. It has all the extensions described in the `GFM
    documentation`_, except for intra-GitHub links to commits, repositories,
    and issues.

    Note that Markdown-formatted gists and files (including READMEs) on GitHub
    use a slightly different variant of GFM. For that, use
    :class:`mdx_partial_gfm.PartialGithubFlavoredMarkdownExtension`.

    .. _GFM documentation: https://guides.github.com/features/mastering-markdown/
    """

    def extendMarkdown(self, md):
        PartialGithubFlavoredMarkdownExtension.extendMarkdown(self, md)
        Nl2BrExtension().extendMarkdown(md)



# Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

"""
:mod:`gfm.autolink` -- Turn URLs into links
===========================================

The :mod:`gfm.autolink` module provides an extension that turns all raw URLs
into marked-up links.

This is based on the `web-only URL regex`_ by John Gruber (public domain).

This regex seems to line up pretty closely with GitHub's URL matching.
Two cases were identified where they differ. In both cases, the
regex were slightly modified to bring it in line with GitHub's parsing:

* GitHub accepts FTP-protocol URLs;
* GitHub only accepts URLs with protocols or ``www.``, whereas Gruber's regex
  accepts things like ``foo.com/bar``.

Typical usage
-------------

.. testcode::

   import markdown
   from gfm import AutolinkExtension

   print(markdown.markdown("I love this http://example.org/ check it out",
                           extensions=[AutolinkExtension()]))

.. testoutput::

   <p>I love this <a href="http://example.org/">http://example.org/</a> check it out</p>


.. _web-only URL regex: http://daringfireball.net/2010/07/improved_regex_for_matching_urls

"""

import re
import markdown
import xml.etree.ElementTree as etree

URL_RE = (
    r"\b((?:(?i:ftp|https?)://|(?i:www)\d{0,3}[.])(?:[^\s()<>]+|"
    r"\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()"
    r"<>]+\)))*\)|[^\s`!()\[\]{};:" + r"'" + r'".,<>?«»“”‘’]))'
)
PROTOCOL_RE = re.compile(r"^(?i:ftp|https?)://")


# We can't re-use the built-in AutolinkPattern because we need to add protocols
# to links without them.
class AutolinkPattern(markdown.inlinepatterns.Pattern):
    def handleMatch(self, m):
        el = etree.Element("a")

        href = m.group(2)
        if not PROTOCOL_RE.match(href):
            href = "http://%s" % href
        el.set("href", self.unescape(href))

        el.text = markdown.util.AtomicString(m.group(2))
        return el


class AutolinkExtension(markdown.Extension):
    """
    An extension that turns URLs into links.
    """

    def extendMarkdown(self, md):
        md.inlinePatterns.register(AutolinkPattern(URL_RE, md), "gfm-autolink", 100)



# Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
# for details. All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.

"""
:mod:`gfm.automail` -- Turn email addresses into links
======================================================

The :mod:`gfm.automail` module provides an extension that turns all raw email
addresses into marked-up links.

Typical usage
-------------

.. testcode::

   import markdown
   from gfm import AutomailExtension

   print(markdown.markdown("You can mail me at foo@example.org for more info",
                           extensions=[AutomailExtension()]))

.. testoutput::

   <p>You can mail me at <a href="mailto:foo@example.org">foo@example.org</a> for more info</p>

"""

import markdown
import markdown.inlinepatterns
import xml.etree.ElementTree as etree

MAIL_RE = r"\b([a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]+)\b"


# We can't re-use the built-in AutomailPattern because we need to add mailto:.
# We also don't care about HTML-encoding the email.
class AutomailPattern(markdown.inlinepatterns.Pattern):
    def handleMatch(self, m):
        el = etree.Element("a")
        el.set("href", self.unescape("mailto:" + m.group(2)))
        el.text = markdown.util.AtomicString(m.group(2))
        return el


class AutomailExtension(markdown.Extension):
    """
    An extension that turns email addresses into links.
    """

    def extendMarkdown(self, md):
        md.inlinePatterns.register(AutomailPattern(MAIL_RE, md), "gfm-automail", 100)
